import numpy as np
from scipy.stats import loguniform

from astropy.timeseries import LombScargle
from stochastic.processes.noise import ColoredNoise

class example_prior:
    """
    This example class defines an example prior class to handle both evaluations and 
    sampling of the prior
    """

    def sample(self, nsamples = None):
        """
        Function that sample points from the prior
        """
        if nsamples is None:
            nsamples = self.nsamples

        # Evaluate samples:
        beta_samples = np.random.uniform(self.beta1, self.beta2, nsamples)
        sigma_samples = loguniform.rvs(self.sigma1, self.sigma2, size = nsamples)

        # Return them:
        return beta_samples, sigma_samples

    def validate(self, theta):
        """
        This function validates that the set of parameters to evaluate 
        are within the ranges of the prior
        """

        # Extract current parameters to evaluate the priors on:
        beta, sigma = theta

        # Validate the uniform prior:
        if beta <= self.beta1 or beta >= self.beta2:
            return False

        # Validate the loguniform prior:
        if sigma <= self.sigma1 or sigma >= self.sigma2:
            return False

        # If all is good, return a nice True:
        return True

    def evaluate(self, theta):
        """
        Given an input vector, evaluate the prior. In this case, this just returns the 
        priors defined by the hyperparameters. For the uniform prior, the value of the 
        prior doesn't depend on the inputs. For the loguniform, that's note the case.
        """

        # Extract current parameters to evaluate the priors on:
        beta, sigma = theta

        # Return the prior evaluated on theta:
        return self.beta_prior * (self.sigma_factor / sigma)


    def __init__(self, beta1 = 0.1, beta2 = 3., sigma1 = 0.1, sigma2 = 100., nsamples = 100):

        # Define hyperparameters of the prior. First for beta (uniform prior):
        self.beta1 = beta1
        self.beta2 = beta2

        # Value of the prior given hyperparameters:
        self.beta_prior = 1. / (beta2 - beta1)

        # Same for sigma (log-uniform):
        self.sigma1 = sigma1
        self.sigma2 = sigma2

        # Compute factor that will multiply 1/x where x is the input to evaluate the 
        # prior:
        la = np.log(sigma1)
        lb = np.log(sigma2)
        self.sigma_factor = 1./(lb - la)  

        # Define the default number of samples:
        self.nsamples = nsamples

def gen_fake_data(length = 1000):
    beta, sigma = 1., 10.

    cn_simulator = ColoredNoise(beta=beta, t = 10)
    simulated_data = cn_simulator.sample(length - 1)
    simulated_data /= np.sqrt(np.var(simulated_data))
    simulated_data *= sigma
    fout = open('data.dat', 'w')
    for i in range(len(simulated_data)):
        fout.write(str(simulated_data[i])+'\n')
    fout.close()

class example_simulator:
    """
    This example class generates a simulator object that is able to simulate several or 
    single simulations
    """

    def single_simulation(self, parameters):

        # Extract parameters:
        beta, sigma = parameters

        # Generate simulation:
        cn_simulator = ColoredNoise(beta=beta, t = self.tend)
        simulated_data = cn_simulator.sample(self.length-1)

        # Standarize it:
        simulated_data /= np.sqrt(np.var(simulated_data))

        # Return scaled time-series:
        return simulated_data * sigma

    def several_simulations(self, parameters):

        # Extract parameters:
        betas, sigmas = parameters
        nsamples = len(betas)

        # Define array to store simulations:
        simulations = np.zeros([nsamples, self.length])

        # Lazy loop to do several of these; could apply multi-processing:
        for i in range(nsamples):
            simulations[i,:] = self.single_simulation([betas[i], sigmas[i]])

        return simulations


    def __init__(self, tend = 10, length = 1000):
        self.tend = tend
        self.length = length

class example_distance:
    """
    Example class for distance.
    """
    def single_distance(self, simulation):
        """ Given a dataset and a simulation, this function returns the distance 
            between them. This is defined here as the median absolute sum between their 
            power-spectral densities, with a time-array defined outside of this function """
        simulation_power = LombScargle(self.times, simulation).power(self.frequencies)
        return np.sum(np.abs(self.data_power - simulation_power))

    def several_distances(self, simulations):
        """ Same as single distance, several times """

        nsimulations = simulations.shape[0]
        distances = np.zeros(nsimulations)
        for i in range(nsimulations):
            distances[i] = self.single_distance(simulations[i,:])
        return distances

    def __init__(self, data, length = 1000):
        self.times = np.linspace(0.01,10.,length)

        self.min_tscale, self.max_tscale = np.median(np.abs(np.diff(self.times))), \
                                           np.max(self.times) - np.min(self.times)

        self.frequencies = np.linspace(1. / self.max_tscale, \
                                       1. / (2. * self.min_tscale), length)

        self.data_power = LombScargle(self.times, data).power(self.frequencies)



import numpy as np
from scipy.stats import multivariate_normal

def get_weights(prior, old_weights, old_thetas, old_covariance, new_thetas, N):
    """ Function that calculates the (normalized) weights of each particle """

    weights = np.zeros(N)
    for j in range(N):
        
        theta = np.array([parameter[j] for parameter in new_thetas])
        
        # Evaluate numerator of equation (3) in Ishida et al. (2015):
        numerator = prior.evaluate(theta)

        # Evaluate denominator:
        denominator = 0.
        for i in range(N):
            old_theta = np.array([parameter[i] for parameter in old_thetas])
            denominator += old_weights[i] * multivariate_normal.pdf(theta, \
                                                                    mean = old_theta,\
                                                                    cov = old_covariance)
        
        # Save weight:
        weights[j] = numerator / denominator

    return weights/np.sum(weights)

def sample(prior, distance, simulator, M = 300, N = 30, Delta = 0.01):
    """
    Given a dataset `data`, a function that return parameters sampled from the prior `prior_sampler`, a function 
    that runs simulations given parameters `simulator` and a `distance` function that given two datasets 
    in the format of `data` returns back the distance between two datasets, this function samples from the 
    posterior using the PMC-ABC algorithm outlined in Ishida et al. (2015; https://arxiv.org/abs/1504.06129).

    Attributes
    ----------
    prior : object
        This is an object that has to have three possible functions. (1) A function `prior.sample(nsamples)` to sample from 
        the prior; this needs to return an N-tuple where each element is an array of length `nsamples` storing the samples 
        from the prior. (2) A function `prior.validate(theta)` that validates `theta` falls within the prior range; if 
        `theta` is within that, this has to return `True` and `False` otherwise. (3) A function `prior.evaluate(theta)` 
        which evaluates the prior at a given parameter vector `theta`.
    distance : object
        Object that calculates the distance from simulations to a(n already loaded) dataset. Has to have two functions; 
        `distance.single_distance` which, given a simulation, returns the distance to that simulation and `distance.several_distances`, 
        which calculates distances to `simulations.shape[0]` simulations stored in an array `simulations`.
    simulator : object
        Object that generates simulations. Must have a `simulator.single_simulation(parameter)` that generates a 
        simulation given a parameter vector and a `simulator.several_simulations(parameter)` that receives a N-tuple, 
        representing each of the parameters, and each tuple is an array, whose i-th element is the parameter for 
        the i-th particle. 
    M : int
        Number of "particles" --- i.e., draws from the prior drawn on the very first iteration of the ABC sampling scheme.
    N : int
        Number of particles on each particle system (i.e., number of particles to keep on that first iteration and in 
        each of the subsequent importance sampling iterations). N must be less than M.
    Delta : float
        Convergence criterion for the ABC sampling scheme: minimum ratio between N and the number of draws necessary to 
        construct a particle system, K. 

    """

    # Check weird user inputs:
    if N > M:
        raise Exception('N ('+str(N)+', number of particles) cannot be larger than M ('+str(M)+', number of each sub particle system). ')

    ######################################################################
    # STEP 1: find the best particles on the initial draw from the prior #
    ######################################################################
    print('\t >> 1. Starting ABC sampler...')

    # First, generate set of samples from the prior:
    thetas = prior.sample(nsamples = M)
    nparameters = len(thetas)

    # Simulate initial particle system:
    simulations = simulator.several_simulations(thetas)

    # Get distances between the dataset and the simulations; sort them out from best to worst, select best N ones:
    distances = distance.several_distances(simulations)
    idx = np.argsort(distances)[:N]
     
    # Save the N best ones to S0 (i.e., S at t=0):
    t = 0
    S = {}
    S[t] = {}
    S[t]['thetas'] = [parameter[idx] for parameter in thetas]
    S[t]['distances'] = distances[idx]

    # Get covariance matrix for these thetas:
    S[t]['covariance'] = np.cov(S[0]['thetas'])

    # Calculate and save inital weights:
    S[t]['weights'] = np.ones(N)/N

    print('\t \t Initial N particles successfully generated.')

    ######################################################################
    # STEP 2: iterate to find sub-particle systems of draws for t > 0    #
    ######################################################################

    # Define parameters that are going to be used in the iteration:
    Current_Delta = np.inf
    idx_quantile = int(N * 0.75)
    idx_N = np.arange(N)

    # Start iteration:
    print('\t >> 2. Going to iterative importance sampling. Trying to reach ', Delta)
    while Current_Delta > Delta:

        # Initialize parameters for the current iteration:
        K = 0
        Kstar = 0
        t = t + 1
        S[t] = {}

        # Calculate epsilon as the 75th quantile of distances in S[t-1]:
        epsilon = S[t-1]['distances'][idx_quantile]

        # Sample N new particles:
        counter = 0
        while True:

            Kstar += 1

            # Sample a proposed theta:
            idx_0 = np.random.choice(idx_N, p = S[t-1]['weights'])
            theta_0 = np.array([parameter[idx_0] for parameter in S[t-1]['thetas']])
            current_theta = np.random.multivariate_normal(theta_0, S[t-1]['covariance'])

            # Before running the simulation, validate that the sampled theta falls 
            # within the bounds of the prior:
            if prior.validate(current_theta):
                current_simulation = simulator.single_simulation(current_theta)
                current_distance = distance.single_distance(current_simulation)
            else:
                current_distance = np.inf

            # If it does, save the sampled theta:
            if current_distance < epsilon:
                if counter == 0:
                    thetas = np.copy(current_theta)
                    distances = np.array([current_distance])
                    Kstars = np.array([Kstar])
                else:
                    thetas = np.vstack((thetas, current_theta))
                    distances = np.append(distances, current_distance)
                    Kstars = np.append(Kstars, Kstar)
                K += Kstar
                Kstar = 0
                counter += 1

            if counter == N:
                break

        # Save the N new particles in the S dictionary. First, save the ordered distances 
        # (this ordering is needed on the next iteration):
        idx = np.argsort(distances)
        S[t]['distances'] = np.copy(distances[idx])

        # Save the ordered thetas:
        thetas = tuple(thetas.T)
        S[t]['thetas'] = [parameter[idx] for parameter in thetas]

        # Get weights:
        S[t]['weights'] = get_weights(prior, S[t-1]['weights'], S[t-1]['thetas'], \
                                      S[t-1]['covariance'], S[t]['thetas'], N)

        # Get (weighted) covariance matrix:
        S[t]['covariance'] = np.cov(S[t]['thetas'], aweights = S[t]['weights'])

        # Calculate the current delta to see if we've completed the sampling:
        Current_Delta = np.double(N)/np.double(K)

        print('\t    At t = ',t,', current Delta is ', Current_Delta,' | Target Delta : ',Delta)

    return S



import re
import numpy
#from distutils.core import setup, Extension
from setuptools import setup, Extension

VERSIONFILE='src/_version.py'
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(name='abeec',
      version=verstr,
      description='abeec: a library to perform approximate bayesian computation',
      url='http://github.com/nespinoza/abec',
      author='Nestor Espinoza',
      author_email='nespinoza@stsci.edu',
      license='MIT',
      packages=['abeec'],
      package_dir={'abeec': 'src'},
      install_requires=['numpy','scipy'],
      python_requires='>=3.0',
      zip_safe=False)



import numpy as np

import abeec 
from utils1 import example_prior, example_distance, example_simulator

# Read simulated dataset:
data = np.loadtxt('data1.dat', unpack = True)

# Perform ABC sampling:
samples = abeec.sampler.sample(example_prior(), example_distance(data), example_simulator())
